#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pyvista as pv
from scipy.interpolate import CubicHermiteSpline
from sklearn.neighbors import KDTree
from tqdm import tqdm

from fibermat import Mat, Net, Stack, Mesh


def vtk_fiber(length=25., width=1., thickness=1., x=0., y=0., z=0.,
              u=1., v=0., w=0., shear=1., tensile=np.inf, index=None,
              r_resolution=1, theta_resolution=8, z_resolution=20, **kwargs):
    """
    Export a fiber as VTK mesh using `pyvista.CylinderStructured <https://docs.pyvista.org/version/stable/api/utilities/_autosummary/pyvista.CylinderStructured.html>`_.

    Parameters
    ----------
    length : float, optional
        Fiber length (mm). Default is 25 mm.
    width : float, optional
        Fiber width (mm). Default is 1 mm.
    thickness : float, optional
        Fiber thickness (mm). Default is 1 mm.
    x : float, optional
        Fiber position: X-coordinate (mm). Default is 0 mm.
    y : float, optional
        Fiber position: Y-coordinate (mm). Default is 0 mm.
    z : float, optional
        Fiber position: Z-coordinate (mm). Default is 0 mm.
    u : float, optional
        Fiber orientation: X-component. Default is 1.
    v : float, optional
        Fiber orientation: Y-component. Default is 0.
    w : float, optional
        Fiber orientation: Z-component. Default is 0.
    shear : float, optional
        Shear modulus (MPa). Default is 1 MPa.
    tensile : float, optional
        Tensile modulus (MPa). Default is ∞ MPa.
    index : int, optional
        Fiber label.

    Returns
    -------
    pyvista.StructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    r_resolution : int, optional
        Number of elements along the radius of the fiber. Default is 1.
    theta_resolution : int, optional
        Number of points on the circular face of the fiber. Default is 8.
    z_resolution : int, optional
        Number of points along the length of the fiber. Default is 20.
    kwargs :
        Additional keyword arguments ignored by the function.

    .. NOTE::
        If `index` is not None, the following fields are added to the VTK mesh:
            - "fiber" : fiber index
            - "lbh" : fiber dimensions (mm)
            - "xyz" : local fiber coordinates (mm)
            - "uvw" : fiber orientation vector
            - "G" : shear modulus (MPa)
            - "E" : tensile modulus (MPa)

    """
    # Create the VTK mesh (cylindrical structured grid)
    vtk = pv.CylinderStructured(radius=np.linspace(0.5, 0, r_resolution + 1),
                                theta_resolution=theta_resolution,
                                z_resolution=z_resolution)

    l, b, h = length, width, thickness

    # Add fields to mesh data
    if index is not None:
        vtk["fiber"] = np.full(len(vtk.points), index)
        vtk["lbh"] = np.tile([l, b, h], (len(vtk.points), 1))
        vtk["xyz"] = vtk.points * np.array([[l, b, h]])
        vtk["uvw"] = np.tile([u, v, w], (len(vtk.points), 1))
        vtk["G"] = np.full(len(vtk.points), shear)
        vtk["E"] = np.full(len(vtk.points), tensile)

    # Transform the mesh (scale, rotate, and translate)
    vtk.scale([l, b, h], inplace=True)
    pv.translate(vtk,
                 center=(x, y, z),
                 direction=(u, v, w))

    # Return VTK mesh
    return vtk


def vtk_mat(mat=None, **kwargs):
    """
    Export a :class:`Mat` object as VTK mesh.

    Parameters
    ----------
    mat : pandas.DataFrame, optional
        Set of fibers represented by a :class:`Mat` object.

    Returns
    -------
    pyvista.UnstructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    kwargs :
        Additional keyword arguments passed to :meth:`vtk_fiber` function.

    .. NOTE::
        The following fields are added to the VTK mesh:
            - "fiber" : fiber index
            - "lbh" : fiber dimensions (mm)
            - "xyz" : local fiber coordinates (mm)
            - "uvw" : fiber orientation vector
            - "G" : shear modulus (MPa)
            - "E" : tensile modulus (MPa)

    """
    # Optional
    mat = Mat.check(mat)

    fibers = []  # : list to store individual fiber meshes

    for i in tqdm(mat.index, desc="Create VTK mat"):
        # Get fiber
        fiber = mat.loc[i].astype(float)
        # Create the VTK mesh (cylindrical structured grid)
        fiber_mesh = vtk_fiber(*fiber[[*"lbhxyzuvwGE"]].values,
                               index=i,
                               **kwargs)
        # Append fiber mesh to list
        fibers.append(fiber_mesh)

    # Combine all individual fiber meshes into a single VTK mesh
    return pv.MultiBlock(fibers).combine()


def vtk_mesh(mat=None, mesh=None, displacement=None, rotation=None,
             force=None, moment=None, **kwargs):
    """
    Export a :class:`Mesh` object as VTK mesh.

    Parameters
    ----------
    mat : pandas.DataFrame, optional
        Set of fibers represented by a :class:`Mat` object.
    mesh : pandas.DataFrame, optional
        Fiber mesh represented by a :class:`Mesh` object.

    Returns
    -------
    pyvista.UnstructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    displacement : numpy.ndarray, optional
        Displacement field.
    rotation : numpy.ndarray, optional
        Rotation field.
    force : numpy.ndarray, optional
        Load field.
    moment : numpy.ndarray, optional
        Torque field.
    kwargs :
        Additional keyword arguments passed to :meth:`vtk_fiber` function.

    .. HINT::
        The following fields are added to the VTK mesh:
            - "fiber" : fiber index
            - "lbh" : fiber dimensions (mm)
            - "xyz" : local fiber coordinates (mm)
            - "uvw" : fiber orientation vector
            - "G" : shear modulus (MPa)
            - "E" : tensile modulus (MPa)
        If `displacement` is not None:
            - "displacement" : displacement field (mm)
            - "rotation" : rotation field (rad)
            - "curvature" : curvature field (1 / mm)
        If `force` is not None:
            - "force" : force field (N)

    """
    # Optional
    mat = Mat.check(mat)
    mesh = Mesh.check(mesh)

    fibers = []  # : list to store individual fiber meshes
    by_fiber = mesh.groupby("fiber")

    for i in tqdm(mat.index, desc="Create VTK mat"):
        # Get fiber
        fiber = mat.loc[i].astype(float)
        # Create the VTK mesh (cylindrical structured grid)
        fiber_mesh = vtk_fiber(*fiber[[*"lbhxyzuvwGE"]].values,
                               index=i,
                               **kwargs)
        # Append fiber mesh to list
        fibers.append(fiber_mesh)

        # Prepare interpolation data
        fiber = by_fiber.get_group(i)
        s = fiber.s.values[:, None]
        x = fiber_mesh["xyz"][:, [0]] * 0.9999
        k = KDTree(s).query(x, return_distance=False).ravel()
        s, x = s.ravel(), x.ravel()
        # Correct indices (s_k <= x_i < s_{k+1}, k \in [-1, n])
        k = k * (x > s[k]) + (k - 1) * (x <= s[k])
        # Indices containing relative distances (np.floor(j) = k)
        j = ((x - s[k]) / (s[k + 1] - s[k]))
        # Correct issues for periodic mesh
        j = np.array([*j])
        j[j == np.inf] = 0
        # Add to vtk_fiber
        j += fiber.index[k]
        fiber_mesh["node"] = j

    # Combine all individual fiber meshes into a single VTK mesh
    vtk = pv.MultiBlock(fibers).combine()

    if len(mat):
        # Interpolate fields
        s = np.arange(len(mesh))
        x = vtk["node"]
        if displacement is not None:
            if rotation is None:
                rotation = 0 * displacement
            displacement = CubicHermiteSpline(s, displacement, rotation)
            vtk["displacement"] = np.zeros(vtk.points.shape)
            vtk["displacement"][:, 2] = displacement(x)
            vtk["rotation"] = displacement.derivative()(x)
            vtk["curvature"] = displacement.derivative(2)(x)
            vtk.points += vtk["displacement"]
        if force is not None:
            if moment is None:
                moment = 0 * force
            force = CubicHermiteSpline(s, force, moment)
            vtk["force"] = force(x)

    # Periodic boundary conditions (optional)
    if len(mat) and mesh.attrs["periodic"]:
        X = Y = mat.attrs["size"]
        Z1, Z2 = np.min(vtk.points), np.max(vtk.points)
        # Duplicate mesh for periodic conditions
        vtk = pv.MultiBlock([
            vtk,
            vtk.copy().translate([-X, 0, 0]),
            vtk.copy().translate([X, 0, 0]),
            vtk.copy().translate([0, -Y, 0]),
            vtk.copy().translate([0, Y, 0]),
            vtk.copy().translate([-X, -Y, 0]),
            vtk.copy().translate([-X, Y, 0]),
            vtk.copy().translate([X, -Y, 0]),
            vtk.copy().translate([X, Y, 0]),
        ]).combine().clip_box([-X, X, -Y, Y, Z1, Z2], invert=False)

    # Return VTK mesh
    return vtk


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    from matplotlib import pyplot as plt

    from fibermat import *

    # Create a VTK fiber
    vtk_fiber().plot()

    # Generate a set of fibers
    mat = Mat(100)
    # Build the fiber network
    net = Net(mat, periodic=True)
    # Stack fibers
    stack = Stack(mat, net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Create a VTK mat
    vtk_mat(mat).plot()

    # Create a VTK mesh
    vtk_mesh(mat, mesh).plot()

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mat, mesh, packing=4, lmin=0.01, coupling=0.99
    )

    # Export as VTK
    vtk = vtk_mesh(mat, mesh,
                   *u(1).reshape(-1, 2).T,
                   *(f(1) @ C).reshape(-1, 2).T)
    vtk.plot(scalars="force", cmap=plt.cm.twilight_shifted)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp
import warnings
from tqdm import tqdm

from fibermat import Mat, Net, Stack, Mesh, stiffness, constraint, Interpolate


def solve(mat, mesh, packing=1., solve=sp.sparse.linalg.spsolve,
           itermax=1000, tol=1e-6, errtol=1e-6, interp_size=None,
           verbose=True, **kwargs):
    r"""An iterative mechanical solver for fiber packing problems. It solves the *quadratic programming problem*:

    .. MATH::
        \min_{\mathbf{u}, \mathbf{f}} \left( \frac{1}{2} \, \mathbf{u} \, \mathbb{K} \, \mathbf{u} - \mathbf{F} \, \mathbf{u} - \mathbf{f} \, (\mathbf{H} - \mathbb{C} \, \mathbf{u}) \right)
    .. MATH::
        \quad s.t. \quad \mathbb{C} \, \mathbf{u} \leq \mathbf{H} \, ,
        \quad \mathbf{u} \leq 0 \, ,
        \quad \mathbf{f} \geq 0
        \quad and \quad \mathbf{f} \, (\mathbf{H} - \mathbb{C} \, \mathbf{u}) = 0

    where:
        - :math:`\mathbf{u}` is the vector of generalized displacements (*unknowns of the problem*).
        - :math:`\mathbf{f}` is the vector of generalized forces (*unknowns Lagrange multipliers*).
        - :math:`\mathbb{K}` is the stiffness matrix of the fiber set.
        - :math:`\mathbf{F}` is the vector of external efforts.
        - :math:`\mathbb{C}` is the matrix of non-penetration constraints.
        - :math:`\mathbf{H}` is the vector of minimal distances between fibers (minimal distances).

    The *mechanical equilibrium* allows reformulating the problem as a system of inequalities:

    .. MATH::
        \Rightarrow \quad \left[\begin{matrix}
            \mathbb{K} & \mathbb{C}^T \\
            \mathbb{C} & 0
        \end{matrix}\right] \binom{\mathbf{u}}{\mathbf{f}}
            \leq \binom{\mathbf{F}}{\mathbf{H}}

    which is solved using an iterative *Updated Lagrangian Approach*.

    .. HINT::
        Models used to build the matrices are implemented in :ref:`🔧 Model`:
            - 𝕂 and 𝑭 : :func:`~.model.stiffness`.
            - ℂ and 𝑯 : :func:`~.model.constraint`.

    Parameters
    ----------
    mat : pandas.DataFrame
        Set of fibers represented by a :class:`Mat` object.
    mesh : pandas.DataFrame
        Fiber mesh represented by a :class:`Mesh` object.
    packing : float, optional
        Targeted value of packing. Must be greater than 1. Default is 1.0.

    Returns
    -------
    tuple
        K : sparse matrix
            Stiffness matrix (symmetric positive-semi definite).
        C : sparse matrix
            Constraint matrix.
        u : Interpolate
            Displacement vector.
        f : Interpolate
            Force vector.
        F : Interpolate
            Load vector.
        H : Interpolate
            Upper-bound vector.
        Z : Interpolate
            Upper-boundary position.
        rlambda : Interpolate
            Compaction.
        mask : Interpolate
            Active rows and columns in the system of inequations.
        err : Interpolate
            Numerical error of the linear solver.

    .. SEEALSO::
        Simulation results are given as functions of a pseudo-time parameter (between 0 and 1) using :class:`~.interpolation.Interpolate` objects.

    Other Parameters
    ----------------
    solve : callable, optional
        Sparse solver. It is a callable object that takes as inputs a sparse symmetric matrix 𝔸 and a vector 𝒃 and returns the solution 𝒙 of the linear system: 𝔸 𝒙 = 𝒃. Default is `scipy.sparse.linalg.spsolve`.
    itermax : int, optional
        Maximum number of solver iterations. Default is 1000.
    tol : float, optional
        Tolerance used for contact detection (mm). Default is 1e-6 mm.
    errtol : float, optional
        Tolerance for the numerical error. Default is 1e-6.
    interp_size : int, optional
        Size of array used for interpolation. Default is None.
    verbose : bool, optional
        If True, displays a progress bar during simulation. Default is True.
    kwargs :
        Additional keyword arguments passed to matrix constructors.

    """
    # Assemble the quadratic programming system
    # TODO: pass a model as argument instead
    K, u, F, du, dF = stiffness(mat, mesh, **kwargs)
    C, f, H, df, dH = constraint(mat, mesh, **kwargs)
    P = sp.sparse.bmat([[K, C.T], [C, None]], format='csc')
    x = np.r_[u, f]
    q = np.r_[F, H]
    dx = np.r_[du, df]
    dq = np.r_[dF, dH]

    u, f = np.split(x, [K.shape[0]])  # Memory-shared
    F, H = np.split(q, [K.shape[0]])  # Memory-shared
    du, df = np.split(dx, [K.shape[0]])  # Memory-shared
    dF, dH = np.split(dq, [K.shape[0]])  # Memory-shared

    u_ = [u.copy()]
    f_ = [f.copy()]
    F_ = [F.copy()]
    H_ = [H.copy()]
    Z_ = [(mesh.z.values + 0.5 * mesh.h.values).max()]
    rlambda_ = [1.0]
    mask_ = [(np.real(q - P @ x) <= tol)]
    err_ = [0]

    # Incremental solver
    with tqdm(total=itermax, desc="Packing: {:.2}".format(rlambda_[-1]),
              disable=not verbose) as pbar:
        i = 0
        while (i < pbar.total) and (rlambda_[-1] < packing):
            # Solve step
            dx *= 0
            # mask = [True] * K.shape[0] + [*(C @ u >= H - tol)]
            mask = (np.real(q - P @ x) <= tol)
            mask &= np.array(np.sum(np.abs(np.real(P)), axis=0) > 0).ravel()
            # Solve linear problem
            sol = solve(P[np.ix_(mask, mask)], dq[mask])
            dx[mask] += np.real(sol)
            # Calculate error
            err = np.linalg.norm(P[np.ix_(mask, mask)] @ dx[mask] - dq[mask])
            # Calculate evolution
            d = np.real(q - P @ x)
            v = np.real(dq - P @ dx)

            try:
                # Calculate the next step
                dU = -min(d[(d > tol) & (v > 0)] / v[(d > tol) & (v > 0)])
                # Stopping criteria
                stop = False
                if err > errtol:
                    if verbose:
                        warnings.warn("Stopping criteria: err = {}".format(err),
                                      UserWarning)
                    stop = True
                if stop:
                    raise ValueError
            except ValueError:
                break

            # Jump to the next step
            x += dx * dU
            q += dq * dU
            # Store results
            u_ += [u.copy()]
            f_ += [f.copy()]
            F_ += [F.copy()]
            H_ += [H.copy()]
            Z_ += [Z_[-1] + dU]
            rlambda_ += [Z_[0] / Z_[-1]]
            mask_ += [mask.copy()]
            err_ += [err.copy()]

            # Update
            i += 1
            pbar.set_description("Packing: {:.2}".format(rlambda_[-1]))
            pbar.update()

    # Interpolate results
    with warnings.catch_warnings():
        # Ignore warning messages due to infinite values in 𝑯
        warnings.filterwarnings('ignore')
        u = Interpolate(u_, size=interp_size)
        f = Interpolate(f_, size=interp_size)
        F = Interpolate(F_, size=interp_size)
        H = Interpolate(H_, size=interp_size)
        Z = Interpolate(Z_, size=interp_size)
        rlambda = Interpolate(rlambda_)
        mask = Interpolate(mask_, kind='previous')
        err = Interpolate(err_, kind='previous')

    # Return interpolated results
    return K, C, u, f, F, H, Z, rlambda, mask, err


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    from fibermat import *

    # Generate a set of fibers
    mat = Mat(100)
    # Build the fiber network
    net = Net(mat)
    # Stack fibers
    stack = Stack(mat, net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mat, mesh, packing=4, lmin=0.01, coupling=0.99
    )

    # Deform the mesh
    mesh.z += u(1)[::2]

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mesh):
        # Draw elements
        for i, j, k in tqdm(zip(mesh.index, mesh.beam, mesh.constraint),
                            total=len(mesh)):
            # Get element data
            a, b, c = mesh.iloc[[i, j, k]][[*"xyz"]].values
            if mesh.iloc[i].s < mesh.iloc[j].s:
                # Draw intra-fiber connection
                plt.plot(*np.c_[a, b],
                         c=plt.cm.tab10(mesh.fiber.iloc[i] % 10))
            if mesh.iloc[i].z < mesh.iloc[k].z:
                # Draw inter-fiber connection
                plt.plot(*np.c_[a, c], '--ok',
                         lw=1, mfc='none', ms=3, alpha=0.2)
            if mesh.iloc[i].fiber == mesh.iloc[k].fiber:
                # Draw fiber end nodes
                plt.plot(*np.c_[a, c], '+k', ms=3, alpha=0.2)
    # Set drawing box dimensions
    ax.set_xlim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    ax.set_ylim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    plt.show()
